#include "DLMZone.h"
#include <cstring>

DLMZone::DLMZone() {
	this->bImageReady = false;
}

DLMZone::DLMZone( int zid, string lid, string type, DLZone * zone, bool brle) {
	
	this->bImageReady = false;
	
	this->bRLE = brle;
	
	this->zid = zid;
	this->lid = lid;
	this->type = type;	
	this->zones.push_back( zone );
	this->rect = this->calRect( zone );
}

DLMZone::~DLMZone() {	
}

/*
DLMZone& DLMZone::operator=( const DLMZone& zone ) {
	this->zid = zone.getID();
	this->lid = zone.getLineID();
	this->type = zone.getType();
	this->rect = zone.getRect();
	this->zones = zone.getZoneList();
	
	return *this;
}
*/
int DLMZone::getID() const {
	return this->zid;
}

string DLMZone::getLineID() const {
	return this->lid;
}

string DLMZone::getType() const {
	return this->type;
}

DLRectShape DLMZone::getRect() const {
	return this->rect;
}

vector<DLZone*> DLMZone::getZoneList() const {
	return this->zones;
}

string DLMZone::getZoneListStr() const {
	
	string str = "";
	
	for( size_t i = 0; i < this->getZoneCnt(); i++ ) {
		str = str + this->zones[i]->dlGetZoneID() + ", ";
	}	
	
	return str.substr(0, str.length()-2);
	
}

DLZone* DLMZone::getZone( int i ) {
	return this->zones[i];
}

void DLMZone::addZone( DLZone * zone ) {
	// rect expansion
	int col, row;	
	
	DLRectShape rect = this->calRect( zone );
	
	col = ( this->rect.dlGetX() < rect.dlGetX() ) ? this->rect.dlGetX() : rect.dlGetX();
	row = ( this->rect.dlGetY() < rect.dlGetY() ) ? this->rect.dlGetY() : rect.dlGetY();
	
	int x1, y1;
	x1 = this->rect.dlGetX() + this->rect.dlGetWidth();
	y1 = this->rect.dlGetY() + this->rect.dlGetHeight();
	
	int x2, y2;
	x2 = rect.dlGetX() + rect.dlGetWidth() - 1;
	y2 = rect.dlGetY() + rect.dlGetHeight() - 1;
	
	int w, h;
	w = ( x1 > x2 ) ? x1 - col : x2 - col;
	h = ( y1 > y2 ) ? y1 - row : y2 - row;
	
	this->rect = DLRectShape( col, row, w+1, h+1 );	
	
	// add zone id
	this->zones.push_back( zone );
	
	this->bImageReady = false;
}

void DLMZone::addZone( DLMZone * zone ) {
	
	for( int i = 0; i < zone->getZoneCnt(); i++ ) {
		addZone( zone->getZone(i) );
	}	
}

int DLMZone::getZoneCnt() const {
	return this->zones.size();
}

bool DLMZone::isOverlap( DLRectShape rect ) {
	
	int tw = this->rect.dlGetWidth();
	int th = this->rect.dlGetHeight();
	int rw = rect.dlGetWidth();
	int rh = rect.dlGetHeight();
	if (rw <= 0 || rh <= 0 || tw <= 0 || th <= 0) {
	    return false;
	}
	int tx = this->rect.dlGetX();
	int ty = this->rect.dlGetY();
	int rx = rect.dlGetX();
	int ry = rect.dlGetY();
	rw += rx;
	rh += ry;
	tw += tx;
	th += ty;
	//      overflow || intersect
	return ((rw < rx || rw > tx) &&
		(rh < ry || rh > ty) &&
		(tw < tx || tw > rx) &&
		(th < ty || th > ry));
	
	/*
	int x = this->rect.dlGetX();
	int y = this->rect.dlGetY();
	int w = this->rect.dlGetWidth();
	int h = this->rect.dlGetHeight();
	
	int X = rect.dlGetX();
	int Y = rect.dlGetY();
	int W = rect.dlGetWidth();
	int H = rect.dlGetHeight();;
	
	if ((w | h | W | H) < 0) {
	    // At least one of the dimensions is negative...
	    return false;
	}
	// Note: if any dimension is zero, tests below must return false...
	
	if (X < x || Y < y) {
	    return false;
	}
	w += x;
	W += X;
	if (W <= X) {
	    // X+W overflowed or W was zero, return false if...
	    // either original w or W was zero or
	    // x+w did not overflow or
	    // the overflowed x+w is smaller than the overflowed X+W
	    if (w >= x || W > w) return false;
	} else {
	    // X+W did not overflow and W was not zero, return false if...
	    // original w was zero or
	    // x+w did not overflow and x+w is smaller than X+W
	    if (w >= x && W > w) return false;
	}
	h += y;
	H += Y;
	if (H <= Y) {
	    if (h >= y || H > h) return false;
	} else {
	    if (h >= y && H > h) return false;
	}
	return true;
	*/	
}

void DLMZone::getRLEList( string rle,
		                  vector<int>& row,
		                  vector<int>& col,
		                  vector<int>& len
		                  ) {
	
	int irow, icol, ilen;
	
	vector<int> all;	
	
	char * str = new char[rle.size()+1];	
	strcpy( str, rle.c_str() );	
	
	char * token;	
	token = strtok( str, ",;()");
	
	while ( token != NULL )	{		
		all.push_back( atoi( token ) );							
		token = strtok( NULL, ",;()");		
	}
	
	for( int i = 0; i < all.size(); i++ ) {
		
		if( i % 3 == 0 ) irow = all[i];
		else if( i % 3 == 1 ) icol = all[i];
		else if( i % 3 == 2 ) {
			ilen = all[i] - icol;		
			row.push_back( irow );
			col.push_back( icol );
			len.push_back( ilen );			
		}		
	}
	
	delete[] str;
}

DLImage DLMZone::getZoneImage(){
		
	if( bImageReady ) return this->image;
	
	DLPolygon poly;	
	DLRectShape rect;
	
	int xOffset;
	int yOffset;
	
	int oxOffset = this->rect.dlGetX();
	int oyOffset = this->rect.dlGetY();
	
	DLImage tmpImage = this->zones[0]->dlGetPageImage();	
	
	DLImage image( this->rect.dlGetWidth(), this->rect.dlGetHeight(), tmpImage.dlGetDepth() );	
	
	for( int i = 0; i < this->getZoneCnt(); i++ ) {
		poly = this->calPolygon( zones[i] );		
		rect = poly.getBounds();
		
		xOffset = rect.dlGetX();
		yOffset = rect.dlGetY();
		
		if( this->bRLE && zones[i]->dlIsTagSet("RLEIMAGE") && zones[i]->dlGetTag("RLEIMAGE") != "" ) {
			// RLE tag decoding		
			
			vector<int> lrow;
			vector<int> lcol;
			vector<int> llen;						
			
			getRLEList( zones[i]->dlGetTag("RLEIMAGE"), lrow, lcol, llen );
			
			for( int s = 0; s < lrow.size(); s++ ) {
				int row = lrow[s];
				int col = lcol[s];
				int len = llen[s];
				
				for( int j = 0; j < len; j++ ) {
					//if( tmpImage.dlGetPixelValue(col + j + xOffset, row + yOffset) != 0 ) {
					//	cout << "Not black pixel" << endl;
					//	continue;
					//}
					try {
						// image.dlSetPixel( col+j, row, tmpImage.dlGetPixelValue( col + j + xOffset, row + yOffset ) );
						image.dlSetPixel( col+j+xOffset, row+yOffset, 0 );
					} catch(...) {
						
					}
				}
			}
			
		} else {
			
			for( int col = 0; col < rect.dlGetWidth(); col++ ) {
				for( int row = 0; row < rect.dlGetHeight(); row++ ) {
					if( poly.contains(col + xOffset, row + yOffset)) {
						try {
							image.dlSetPixel( col + xOffset - oxOffset, row + yOffset - oyOffset, tmpImage.dlGetPixelValue( col + xOffset, row + yOffset ) );
						} catch(...) {
							
						}
					}
				}
			}		
		}
	}	
	
	this->bImageReady = true;
	this->image = image;
	
	return image;
/*	
	vector<DLZone*>::iterator eItr = this->zones.end();
	for( vector<DLZone*>::iterator itr = this->zones.begin();
	     itr != eItr;
	     itr++ ) {
		
		poly = this->calPolygon( *itr );
		rect = poly.getBounds();		
		
		if( (*itr)->dlIsTagSet("RLEIMAGE") ) {
			// RLE tag decoding		
			
			vector<int> lrow;
			vector<int> lcol;
			vector<int> llen;						
			
			getRLEList( (*itr)->dlGetTag("RLEIMAGE"), lrow, lcol, llen );
			
			for( int i = 0; i < lrow.size(); i++ ) {
				int row = lrow[i];
				int col = lcol[i];
				int len = llen[i];
				
				for( int j = 0; j < len; j++ ) {
					image.dlSetPixel( col+j, row, tmpImage.dlGetPixelValue( rect.dlGetX() + col + j, rect.dlGetY() + row ) );
				}
			}
			
		} else {
			
			for( int col = rect.dlGetX(); col < rect.dlGetX() + rect.dlGetWidth(); col++ ) {
				for( int row = rect.dlGetY(); row < rect.dlGetY() + rect.dlGetHeight(); row++ ) {
					if( poly.contains(col, row)) {
						try {
							image.dlSetPixel( xDiff + col, yDiff + row, tmpImage.dlGetPixelValue( col, row ) );
						} catch(...) {
							
						}
					}
				}
			}		
		}
	}
	*/		
}

string DLMZone::getRunLengthCode() {
	string str = "";
	bool bStart;
	int iCol;
	
	DLImage image = this->getZoneImage();
	
	for( int row = 0; row < image.dlGetHeight(); row++ ) {
		bStart = false;
		for( int col = 0; col < image.dlGetWidth(); col++ ) {
			if( image.dlGetPixelValue(col, row) == 0 && bStart == false ) {
				bStart = true;
				iCol = col;
			} else if( image.dlGetPixelValue(col, row) != 0 && bStart == true ) {
				bStart = false;
				str = str + "(" + cvIntToStr(row) + "," + cvIntToStr(iCol) + "," + cvIntToStr(col) + ");";
			}
		}
	}	
	
	return str;
}

string DLMZone::cvIntToStr(int num) {
	
	char buf[10];
	sprintf( buf, "%d", num);
	return string(buf);
}

double DLMZone::getCoverScore( DLMZone & zone ) {
	
	int cntInterRect = this->cntOverlapPixel( zone );	
	return ( (double)cntInterRect / zone.cntONPixel() ) * 100;
}

double DLMZone::getMatchScore( DLMZone & zone ) {
	
	int	cntInterRect = this->cntOverlapPixel( zone );
	/*
	int temp = this->cntONPixel();
	int itemp = zone.cntONPixel();
	
	return ( (double)cntInterRect / ( (this->cntONPixel() + zone.cntONPixel()) - cntInterRect ) ) * 100;
	*/
	double precision = (double)cntInterRect / zone.cntONPixel() * 100;
	double recall = (double)cntInterRect / this->cntONPixel() * 100;
	
	return 2 * precision * recall / ( precision + recall );
	
}

int DLMZone::cntONPixel() {
	
	DLImage image = this->getZoneImage();
	int cnt = 0;

	for( int col = 0; col < image.dlGetWidth(); col++ ) {
		for( int row = 0; row < image.dlGetHeight(); row++ ) {
			if( image.dlGetPixelValue(col, row) != 255 ) cnt++;
		}
	}	
	return cnt;	
}

int DLMZone::cntOverlapPixel( DLMZone& zone ) {
	
	DLImage image = this->getZoneImage();
	DLImage destImage = zone.getZoneImage();
	
	//DLImageFactory::dlGetInstance()->dlSaveImage( image, "tstImages.tif" );
	//DLImageFactory::dlGetInstance()->dlSaveImage( destImage, "tstImaged.tif");
	
	DLRectShape interRect = this->getInterRect( this->rect, zone.getRect() );	
	
	int cnt = 0;	
	
	int sxDiff = interRect.dlGetX() - this->rect.dlGetX();
	int syDiff = interRect.dlGetY() - this->rect.dlGetY();
	int dxDiff = interRect.dlGetX() - zone.getRect().dlGetX();
	int dyDiff = interRect.dlGetY() - zone.getRect().dlGetY();
	
	//cout << "X = " << this->rect.dlGetX() << ", Y = " << this->rect.dlGetY() << endl;
	//cout << "W = " << image.dlGetWidth() << ", H = " << image.dlGetHeight() << endl;	
	//cout << "diffX = " << sxDiff << ", diffY = " << syDiff << endl;

	//cout << "X = " << zone.getRect().dlGetX() << ", Y = " << zone.getRect().dlGetY() << endl;
	//cout << "W = " << destImage.dlGetWidth() << ", H = " << destImage.dlGetHeight() << endl;	
	//cout << "diffX = " << dxDiff << ", diffY = " << dyDiff << endl;
	
	//cout << "x = " << interRect.dlGetX() << ", y = " << interRect.dlGetY() << endl;
	//cout << "w = " << interRect.dlGetWidth() << ", h = " << interRect.dlGetHeight() << endl;
	
	for( int col = 0; col < interRect.dlGetWidth(); col++ ) {
		for( int row = 0; row < interRect.dlGetHeight(); row++ ) {
			if( image.dlGetPixelValue(col + sxDiff, row + syDiff) != 255 
				&& ( destImage.dlGetPixelValue(col + dxDiff, row + dyDiff) == image.dlGetPixelValue(col + sxDiff, row + syDiff) ) )
				cnt++;
		}
	}	
	return cnt;
}

DLRectShape DLMZone::calRect( const DLZone * zone ) {	
	return calPolygon( zone ).getBounds();
}

DLPolygon DLMZone::calPolygon( const DLZone * zone ) {
	
	DLPolygon poly;
		
	if( zone->dlIsTagSet("polygon") ) {	
		// polygon shape		
		poly = this->getPolyFromPoly( zone );
	} else if( zone->dlIsTagSet("orientation") || zone->dlIsTagSet("orientationD") ) {
		// oriented box
		poly = this->getPolyFromOri( zone );
	} else {
		// normal box
		poly = this->getPolyFromRect( zone );
	}
	
	return poly;
	
}

DLPolygon DLMZone::getPolyFromPoly( const DLZone * zone ) {

	std::list<int> xlist;
	DLPolygon poly;
		
	int iSize;
	
	int * x;
	int * y;	
		
	string polyStr = zone->dlGetTag("polygon");
	char * str = new char[ polyStr.size()+1];	
	strcpy( str, polyStr.c_str() );	
	
	char * token;	
	token = strtok( str, ",;()");
	
	while ( token != NULL )	{		
		xlist.push_back( atoi( token ) );							
		token = strtok( NULL, ",;()");		
	}
	
	iSize = xlist.size();
	
	x = new int[iSize/2];
	y = new int[iSize/2];
	
	std::list<int>::iterator xIt = xlist.begin();	
	
	for( int i = 0; i < iSize; i++, xIt++ ) {
		i % 2 == 0 ? x[i/2] = *xIt : y[i/2] = *xIt;
	}
	
	poly = DLPolygon( x, y, iSize/2 );
	
	delete[] x;
	delete[] y;
	delete[] str;
	
	return poly;
}

DLPolygon DLMZone::getPolyFromOri( const DLZone * zone ) {
	
	double pi = 3.14159265;
		
	int xpoint[4];
	int ypoint[4];
	
	int iOrgX = zone->dlGetZoneOrigin().x;	
	int iOrgY = zone->dlGetZoneOrigin().y;
	int iOrgW = zone->dlGetZoneWidth();
	int iOrgH = zone->dlGetZoneHeight();
	
	// Radian rotation
	
	double dAngle;
	
	if( zone->dlIsTagSet("orientation")) dAngle = atof(zone->dlGetTag("orientation").c_str());
	else dAngle = pi / 180 * atof(zone->dlGetTag("orientationD").c_str()); 
	
	// cout << "zoneID " << zone.dlGetTag("id") << endl;
	
	xpoint[0] = iOrgX;
	ypoint[0] = iOrgY;
	
	xpoint[1] = iOrgX + transPosX( iOrgW, 0, dAngle );
	ypoint[1] = iOrgY - transPosY( iOrgW, 0, dAngle );
	
	xpoint[2] = iOrgX + transPosX( iOrgW, -iOrgH, dAngle );
	ypoint[2] = iOrgY - transPosY( iOrgW, -iOrgH, dAngle );
	
	xpoint[3] = iOrgX + transPosX( 0, -iOrgH, dAngle );
	ypoint[3] = iOrgY - transPosY( 0, -iOrgH, dAngle );
	
	return DLPolygon( xpoint, ypoint, 4 );	
}

DLPolygon DLMZone::getPolyFromRect( const DLZone * zone ) {
	int xpoint[4];
	int ypoint[4];
	
	int col = zone->dlGetZoneOrigin().x;		
	int row = zone->dlGetZoneOrigin().y;
	int width = zone->dlGetZoneWidth();	
	int height = zone->dlGetZoneHeight();	
	
	xpoint[0] = col;
	ypoint[0] = row;
	
	xpoint[1] = col + width - 1;
	ypoint[1] = row;
	
	xpoint[2] = col + width - 1;
	ypoint[2] = row + height - 1;
	
	xpoint[3] = col;
	ypoint[3] = row + height -1;
	
	return DLPolygon( xpoint, ypoint, 4 );	
}

DLRectShape DLMZone::getInterRect( DLRectShape src, DLRectShape tgt ) {
	
	int tx1 = src.dlGetX();
	int ty1 = src.dlGetY();
	int rx1 = tgt.dlGetX();
	int ry1 = tgt.dlGetY();
	long tx2 = tx1; tx2 += src.dlGetWidth();
	long ty2 = ty1; ty2 += src.dlGetHeight();
	long rx2 = rx1; rx2 += tgt.dlGetWidth();
	long ry2 = ry1; ry2 += tgt.dlGetHeight();
	if (tx1 < rx1) tx1 = rx1;
	if (ty1 < ry1) ty1 = ry1;
	if (tx2 > rx2) tx2 = rx2;
	if (ty2 > ry2) ty2 = ry2;
	tx2 -= tx1;
	ty2 -= ty1;
	
	return DLRectShape(tx1, ty1, (int)tx2, (int)ty2);
}

int DLMZone::transPosX( int x, int y, double angle ) {
	double result = x * cos(angle) - y * sin(angle); 
	return (int)result;
}

int DLMZone::transPosY( int x, int y, double angle ) {
	double result = x * sin(angle) + y * cos(angle); 
	return (int)result;
}


