#include <iostream>
#include <cstring>
#include <time.h>
#include <list>
#include <iomanip>

#include <sys/stat.h>

#include "boost/filesystem/fstream.hpp"
#include "boost/filesystem/operations.hpp"
#include "boost/filesystem/path.hpp"

#include "DLxml.h"

// from Doclib
#include "DLDocument.h"
#include "DLPage.h"
#include "DLZone.h"
#include "DLConst.h"
#include "DLRectShape.h"
#include "DLImageFactory.h"

#include "DLPets.h"
#include "DLPolygon.h"

using namespace std;
namespace fs = boost::filesystem;

bool compare_function( Score first, Score second ) {
	if( first.dlScore < second.dlScore ) return true;
	else return false;
}

DLPets::DLPets() {
	
}

DLPets::DLPets(int argc, char** argv) {
	this->initialize();
	
	for( int i = 0; i < argc; i++ ) 
		this->strOptions = this->strOptions + string( argv[i] ) + " ";	
	
	this->parseOptions(argc, argv);	
}

DLPets::~DLPets() {
	// release memory
	if( this->iConfMatrix ) {
		for( UINT i = 0; i < this->iCntType; i++ ) delete[] iConfMatrix[i];
		delete[] iConfMatrix;
	}
	
	if( this->iMissing ) delete[] this->iMissing;
	if( this->iFalseAlram ) delete[] this->iFalseAlram;
	
	if( this->strZType ) delete[] this->strZType;
}

void DLPets::initialize() {
	//this->gPath = fs::path( fs::initial_path<fs::path>() );
	//this->oPath = fs::path( fs::initial_path<fs::path>() );
	// this->rPath = fs::path( "eval-out.txt" );
	
	this->iFileCnt = 0;
	
	this->strOptions = "";
	
	this->bMultiToMulti = false;
	this->bSingleToMulti = false;
	this->bMultiToSingle = false;
	this->bSingleToSingle = false;
	
	this->bZC = false;
	this->bDetail = false;
	this->bVisual = false;
	this->bMerge = false;
	this->bLineID = false;
	this->bThreshold = false;
	this->bSegOnly = false;
	this->bRunLength = false;
	
	this->bAZ = false;
	this->bNAZ = false;
	
	this->mergeTH = DEFAULT_THRESHOLD;
	
	this->strZType = NULL;
	
	this->iConfMatrix = NULL;
	this->iCntType = 0;
	this->iMissing = NULL;
	this->iFalseAlram = NULL;
	
	this->iCntGZones = 0;
	this->iCntRZones = 0;
	this->iCntSRectZones = 0;
	this->iCntSPolyZones = 0;
	this->iCntSOriZones = 0;
	this->iCntTRectZones = 0;
	this->iCntTPolyZones = 0;
	this->iCntTOriZones = 0;	
}

void DLPets::addType( string type ) {
	
	if( !exist( type ) ) {
	
		string * temp = new string[ this->iCntType+1 ];	
		
		for( size_t i = 0; i < this->iCntType; i++ ) {
			temp[i] = this->strZType[i];
		}
		temp[this->iCntType] = type;

		if( this->strZType ) delete[] this->strZType;
		this->strZType = new string[ this->iCntType+1 ];
		
		// copy
		for( size_t i = 0; i < this->iCntType+1; i++ ) this->strZType[i] = temp[i];
		
		delete[] temp;
		
		UINT ** itemp = new UINT*[this->iCntType+1];
		for( UINT i = 0; i < this->iCntType+1; i++ ) itemp[i] = new UINT[this->iCntType+1];
		for( UINT i = 0; i < this->iCntType+1; i++ )
			for( UINT j = 0; j < this->iCntType+1; j++ ) itemp[i][j] = 0;	
		
		for( UINT i = 0; i < this->iCntType; i++ ) {
			for( UINT j = 0; j < this->iCntType; j++ ) {
				itemp[i][j] = this->iConfMatrix[i][j];
			}		
		}
		
		if( this->iConfMatrix ) {
			for( UINT i = 0; i < iCntType; i++ ) delete[] this->iConfMatrix[i];
			delete[] this->iConfMatrix;
		}
		
		this->iConfMatrix = new UINT*[this->iCntType+1];
		for( UINT i = 0; i < this->iCntType+1; i++ ) this->iConfMatrix[i] = new UINT[this->iCntType+1];
		for( UINT i = 0; i < this->iCntType+1; i++ )
			for( UINT j = 0; j < this->iCntType+1; j++ ) this->iConfMatrix[i][j] = itemp[i][j];
		
		for( UINT i = 0; i < iCntType+1; i++ ) delete[] itemp[i];
		delete[] itemp;		
			
		mapZType[type] = this->iCntType;
		
		this->iCntType++;
	}
}

bool DLPets::exist( string type ) {
	for( UINT i = 0; i < this->iCntType; i++ )
		if( this->strZType[i] == type ) return true;
	
	return false;
}

void DLPets::parseOptions(int argc, char** argv) {
	
	int iPos;
	
	bool bGround = false;
	bool bOutput = false;
	bool bResult = false;
	bool bImage  = false;
	
	for( int i = 1; i < argc; i++ ) {
		if( !strcmp( argv[i],"-g" ) ) {			
			iPos = i;
			bGround = true;
			// this->gPath = fs::system_complete( fs::path( argv[iPos+1], fs::native ) ).normalize();
			this->gPath = fs::path( argv[iPos+1] ).normalize();
			if( !fs::exists( this->gPath ) ) {
				cout << "Ground truth path " << argv[iPos+1] << " is not found!!!" << endl;
				exit(0);
			}
			continue;
		}
		
		if( !strcmp( argv[i], "-r" ) ) {
			iPos = i;
			bOutput = true;			
			// this->oPath = fs::system_complete( fs::path( argv[iPos+1], fs::native ) ).normalize();
			this->oPath = fs::path( argv[iPos+1] ).normalize();
			if( !fs::exists( this->oPath ) ) {
				cout << "Output path " << argv[iPos+1] << " is not found!!!" << endl;
				exit(0);
			}
			continue;
		}
		
		if( !strcmp( argv[i], "-detail" ) ) {
			bDetail = true;			
			continue;
		}
		
		if( !strcmp( argv[i], "-o" ) ) {
			iPos = i;
			bResult = true;
			// this->rPath = fs::system_complete( fs::path( argv[iPos+1], fs::native ) ).normalize();
			this->rPath = fs::path( argv[iPos+1] ).normalize();
			
			continue;
		}
		
		if( !strcmp( argv[i], "-i" ) ) {
			iPos = i;
			bImage = true;
			// this->imgPath = fs::system_complete( fs::path( argv[iPos+1], fs::native ) ).normalize();
			this->imgPath = fs::path( argv[iPos+1] ).normalize();
			if( !fs::exists( this->imgPath ) ) {
				cout << "Image path " << argv[iPos+1] << " is not found!!!" << endl;
				exit(0);
			}
			continue;
		}
		
		if( !strcmp( argv[i], "-v" ) ) {
			iPos = i;
			bVisual = true;
			// this->vPath = fs::system_complete( fs::path( argv[iPos+1], fs::native) ).normalize();
			this->vPath = fs::path( argv[iPos+1] ).normalize();
			if( !fs::exists( this->vPath ) ) {
				// cout << this->vPath.string().c_str() << endl;
				fs::create_directory( this->vPath.string().c_str() );
			}
			continue;
		}
		
		if( !strcmp( argv[i], "-t" ) ) {
			iPos = i;
			this->dMatchTh = atof( argv[iPos+1] );
			bThreshold = true;
			continue;
		}
		
		if( !strcmp( argv[i], "-segonly" ) ) {
			bSegOnly = true;
			this->accZoneList.push_back("zone");
			continue;
		}
		
		if( !strcmp( argv[i], "-zoneclass")) {
			bZC = true;
			continue;
		}
		
		if( !strcmp( argv[i], "-rle" ) ) {
			bRunLength = true;
			continue;
		}
		
		if( !strcmp( argv[i], "-m" ) ) {
			bMerge = true;
			iPos = i;
			
			string instr;
			ifstream infile( argv[iPos+1] );
			if( infile == NULL ) {
				cout << argv[iPos+1] << " is not found" << endl;
				exit(0);
			}
			
			// first line is threshold for merging
			infile >> instr;
			this->mergeTH = atof( instr.c_str() );
			
			while( !infile.eof() ) {
				infile >> instr;
				this->mergeList.push_back( instr );			
			}
			
			infile.close();
			
			continue;
		}
		
		if( !strcmp( argv[i], "-lid" ) ) {
			bLineID = true;
			continue;
		}
		
		if( !strcmp( argv[i], "-az" ) ) {
			bAZ = true;
			iPos = i;
			
			string instr;
			ifstream infile( argv[iPos+1] );
			if( infile == NULL ) {
				cout << argv[iPos+1] << " is not found" << endl;
				exit(0);
			}
												
			while( !infile.eof() ) {
				infile >> instr;
				this->accZoneList.push_back( instr );
			}
			
			infile.close();
			
			continue;
		}
		
		if( !strcmp( argv[i], "-naz" ) ) {
			bNAZ = true;
			iPos = i;
						
			string instr;
			ifstream infile( argv[iPos+1] );
			if( infile == NULL ) {
				cout << argv[iPos+1] << " is not found" << endl;
				exit(0);
			}
												
			while( !infile.eof() ) {
				infile >> instr;
				this->nAccZoneList.push_back( instr );
			}
			
			infile.close();
			
			continue;
		}		
	}
	
	if( bAZ && bNAZ ) {
		cout << "-ac and -nac options can not be used together." << endl;
		exit(0);
	}
	
	if( bSegOnly && bZC ) {
		cout << "-segonly and -zoneclass options can not be used together." << endl;
		exit(0);
	}
	
	if( !bGround ) {
		cout << "This program requires \"-g [DIR | FILE]\" option. " << endl;
		exit(0);
	}
	
	if( !bOutput ) {
		cout << "This program requires \"-r [DIR | FILE]\" option. " << endl;
	}
	
	if( !bResult ) {
		cout << "The Evaluation result will be exported to \"PETS-Eval.txt\"." << endl;
		this->rPath = fs::path( "PETS-Eval.txt" );
	}
	
	if( !bImage ) {
		if( fs::is_directory( this->gPath ) ) this->imgPath = this->gPath;		
		else if( fs::is_regular( this->gPath) ) {
			// todo : Check if it can be given the parent path from the file path.
			cout << "Image path should be provided when Ground truth path is not a directory" << endl;
			exit(0);
		}
	}
	
	if( !bThreshold ) {
		this->dMatchTh = 80.0;
	}	
	
	if( fs::is_directory( this->gPath ) && fs::is_directory( this->oPath ) )
		this->bMultiToMulti = true;	
	if( fs::is_regular( this->gPath ) && fs::is_directory( this->oPath ) )
		this->bSingleToMulti = true;
	if( fs::is_directory( this->gPath ) && fs::is_regular( this->oPath ) )
		this->bMultiToSingle = true;
	if( fs::is_regular( this->gPath ) && fs::is_regular( this->oPath ) )
		this->bSingleToSingle = true;
}

/*
 * 
 * 
 */
void DLPets::compareDoc( const string& src,
		                 const string& target,		                 
		                 const string& save,
		                 ofstream & out
		                 ) {
	
	DLDocument srcDoc;
	DLDocument targetDoc;	
	
	try {		
		srcDoc = (new DLxml())->dlReadFromXmlFile( src, this->imgPath.string() );		
	} catch(...) {
		cout << "File read Error at " << src << endl;
		return;
	}
	
	cout << src << endl;
	
	try {		
		targetDoc = (new DLxml())->dlReadFromXmlFile( target, this->imgPath.string() );
	} catch(...) {
		cout << "File read Error at " << target << endl;
		return;
	}
	
	this->iFileCnt = this->iFileCnt + 1;
	
	this->iMaxZoneNum = 0;
	
	if( srcDoc.dlHasPages() && targetDoc.dlHasPages() ) {		
		
		if( bDetail ) {
			out << srcDoc.dlGetTag("src") << endl;
			for( UINT i = 0; i < srcDoc.dlGetTag("src").size(); i++ ) out << "=";		
			out << endl;
			out << endl;
		}
		
		DLDocument::DLPagePtrIterator tgtPageItr = targetDoc.begin();
		
		DLDocument::DLPagePtrIterator srcEndPItr = srcDoc.end();		
		for( DLDocument::DLPagePtrIterator srcPageItr = srcDoc.begin();
		     srcPageItr != srcEndPItr;
		     srcPageItr++, tgtPageItr++ ) {			
			
			if( bZC )
				this->compareCPage( **srcPageItr, **tgtPageItr, out );
			else
				this->compareLPage( **srcPageItr, **tgtPageItr, out );
		}
		
		if( bVisual ) {
			(new DLxml( srcDoc, srcDoc.dlGetTag("src")))->dlSaveToXmlFile( save );
		}
	}	
}

string DLPets::cvIntToStr( int num ) {
		
	char buf[10];
	sprintf( buf, "%d", num );
	
	return string(buf);	
}

string DLPets::cvDblToStr( double num ) {
	
	char buf[20];
	sprintf( buf, "%.2f", num );
	
	return string(buf);
}

int DLPets::cvStrToInt( string str ) {
	return atoi( str.c_str() );
}

bool DLPets::isMergeList( string type ) {
	
	for( size_t i = 0; i < this->mergeList.size(); i++ ) {
		if( type == this->mergeList[i] ) return true;
	}
	return false;
}

bool DLPets::checkAccZone(string type) {
	if( this->accZoneList.size() == 0 ) return false;
	
	for( size_t i = 0; i < this->accZoneList.size(); i++ ) {
		if( this->accZoneList[i] == type ) return true;
	}
	return false;
}

bool DLPets::checkNotAccZone(string type) {
	if( this->nAccZoneList.size() == 0 ) return false;
	
	for( size_t i = 0; i < this->nAccZoneList.size(); i++ ) {
		if( this->nAccZoneList[i] == type ) return true;
	}
	return false;
}

void DLPets::getMZones( DLPage& page, DLMZone * zones, UINT& size, DLMZone * tZones, UINT tSize ) {
	
	string strZoneID;
	int iID = 0;
	int mid = 0;
	string strLID;	
	string strType;
	
	// clear the garbage zones of the page before merging
	for( DLPage::DLZonePtrIterator zItr = page.begin();
		 zItr != page.end();
		 zItr++ ) {		
		
		strType = (*zItr)->dlGetTag("gedi_type");
		
		// if type is not in the ACCEPTED ZONE LIST, delete the zone from the page
		if( this->bAZ && !checkAccZone( strType ) ) {
			page.dlDeleteZone(zItr);
			zItr--;		// to get the origin position
			continue;
		}
		// if type is in Not ACCEPTED ZONE LIST, delete the zone from the page
		if( this->bNAZ && checkNotAccZone( strType ) ) {
			page.dlDeleteZone(zItr);
			zItr--;
			continue;
		}		
	}
	
	bool ** bCoverTable = new bool*[page.dlGetNumZones()];
	for( int i = 0; i < page.dlGetNumZones(); i++ )
		bCoverTable[i] = new bool[tSize];
		
	DLPage::DLZonePtrIterator zItr = page.begin();
	for( int row = 0; row < page.dlGetNumZones(); row++, zItr++ ) {		
		
		if( this->bSegOnly ) strType = "zone";
		else strType = (*zItr)->dlGetTag("gedi_type");
		
		this->addType( strType );
		
		DLMZone newZone(iID, strLID, strType, *zItr, this->bRunLength);
		
		for( size_t col = 0; col < tSize; col++ ) {
			
			bCoverTable[row][col] = false;			
			
			if( newZone.isOverlap( tZones[col].getRect() ) ) {				
				
				if( isMergeList( newZone.getType() ) 
					&& tZones[col].getCoverScore( newZone ) >= this->mergeTH ) {
					// score matrix update
					bCoverTable[row][col] = true;
				}	
				
			}
		}
	}	
	
	// check the zone to be merged
	for( size_t col = 0; col < tSize; col++ ) {
		std::vector<int> vMList;
		for( size_t row = 0; row < page.dlGetNumZones(); row++ ) {
			if( bCoverTable[row][col] == true ) vMList.push_back(row);
		}
		
		if( vMList.size() == 0 ) continue;
	
		// todo : check the merge list is already exist		
		// todo : treat multiple MZID if one result zone is a member of multi ground truth zone
		
		DLPage::DLZonePtrIterator zoneItr = page.begin();
		for( int cnt = 0; cnt < vMList[0]; cnt++) zoneItr++;
			
		if( vMList.size() == 1 ) {
			strLID = (*zoneItr)->dlGetZoneID();
		} else {
			strLID = "M" + cvIntToStr(mid++);
		}	
		
		if( this->bSegOnly ) strType = "zone";
		else strType = (*zoneItr)->dlGetTag("gedi_type");
		
		zones[iID] = DLMZone( iID, strLID, strType, *zoneItr, this->bRunLength );
		
		(*zoneItr)->dlSetTag("MZID", cvIntToStr(iID), true);
		
		for( size_t i = 1; i < vMList.size(); i++ ) {
			
			zoneItr = page.begin();
			for( int cnt = 0; cnt < vMList[i]; cnt++) zoneItr++;			
			
			zones[iID].addZone( *zoneItr );			
			
			(*zoneItr)->dlSetTag("MZID", cvIntToStr(iID), true);
			
		}
		
		iID++;		
		// cout << "Merged" << endl;
	}
	
	bool bIS;
	// add zone which is not covered by any GT
	DLPage::DLZonePtrIterator znItr = page.begin();
	for( int row = 0; row < page.dlGetNumZones(); row++, znItr++ ) {
		bIS = false;
		
		for( size_t col = 0; col < tSize; col++ ) {
			if( bCoverTable[row][col] == true ) {
				bIS = true;
				break;
			}
		}
		
		if( bIS == false ) {
			strLID = (*znItr)->dlGetZoneID();
			
			if( this->bSegOnly ) strType = "zone";
			else strType = (*znItr)->dlGetTag("gedi_type");
			
			zones[iID] = DLMZone( iID, strLID, strType, *znItr, this->bRunLength);
			(*znItr)->dlSetTag("MZID", cvIntToStr(iID), true);
			
			iID++;
		}
	}
	size = iID;

	for( int i = 0; i < page.dlGetNumZones(); i++ ) delete[] bCoverTable[i];
	delete[] bCoverTable;
	
}

void DLPets::getZones( DLPage& page,
		               DLMZone * zones,
		               UINT& size		     
		               ) {

	string zoneID;
	int id = 0;	
	string lid;
	
	string type;
	
	bool bAdded;	
		
	if( page.dlHasZones() ) {
		
		for( DLPage::DLZonePtrIterator zItr = page.begin();
		     zItr != page.end();
		     zItr++ ) {			
			
			bAdded = false;			
			
			type = (*zItr)->dlGetTag("gedi_type");
			
			// if type is not in the ACCEPTED ZONE LIST, delete the zone from the page
			if( this->bAZ && !checkAccZone( type ) ) {
				page.dlDeleteZone(zItr);
				zItr--;		// to get the origin position
				continue;
			}
			// if type is in Not ACCEPTED ZONE LIST, delete the zone from the page
			if( this->bNAZ && checkNotAccZone( type ) ) {
				page.dlDeleteZone(zItr);
				zItr--;
				continue;
			}
			
			if( this->bSegOnly ) type = "zone";
			else type = (*zItr)->dlGetTag("gedi_type");
			
			this->addType( type );
			
			if( this->bLineID ) {
				if( (*zItr)->dlIsTagSet("lineID") ) {
								
					// check if the zone have a same line id
					lid = (*zItr)->dlGetTag("lineID");

					// Bug possible
					// if there is no match
					for( size_t i = 0; i < size; i++ ) {
						if( lid != "" && zones[i].getLineID() == lid ) {						
							zones[i].addZone( *zItr );
							(*zItr)->dlSetTag("MZID", cvIntToStr( zones[i].getID() ), true );
							bAdded = true;
							break;
						}
					}
					
					if( bAdded == false ) {
					// if same lid is not found
						zones[size++] = DLMZone( id, lid, type, *zItr, this->bRunLength);					
						(*zItr)->dlSetTag("MZID", cvIntToStr(id++), true);
					}
				} else {	// LineID
					// add new zone
					zones[size++] = DLMZone( id, (*zItr)->dlGetZoneID(), type, *zItr, this->bRunLength);
					(*zItr)->dlSetTag("MZID", cvIntToStr(id++), true);
				}
				
			} else {
				zones[size++] = DLMZone( id, (*zItr)->dlGetZoneID(), type, *zItr, this->bRunLength);
				(*zItr)->dlSetTag("MZID", cvIntToStr(id++), true);
			}			
		}
	}
}

/*
void DLPets::setIDType( DLPage& page, string* id, string* type ) {
	
	int i = 0;
	DLPage::DLZonePtrIterator eIter = page.end();
	for( DLPage::DLZonePtrIterator Iter = page.begin();
	     Iter != page.end();
	     Iter++, i++ ) {
		id[i] = (*Iter)->dlGetZoneID();
		type[i] = (*Iter)->dlGetTag("gedi_type");
		// if( !exist( type[i] ) ) this->addType( type[i] );
	}
}
*/

DLZone* DLPets::findZoneByID( DLPage& page, string id ) {
	
	DLZone* zone = NULL;
	
	for( DLPage::DLZonePtrIterator sIter = page.begin();
		 sIter != page.end();
		 sIter++ ) {		
		if( (*sIter)->dlGetZoneID() == id ) {
			zone = *sIter;
			break;
		}
	}
	return zone;
}

void DLPets::compareCPage( DLPage& srcPage, DLPage& tgtPage, ofstream & out ) {
	
	DLZone * zone;
	size_t szZoneTotal = 0;
	size_t szMatchCnt = 0;
	size_t szUnmatchCnt = 0;
	
	this->iCntRZones = this->iCntRZones + srcPage.dlGetNumZones();
	this->iCntGZones = this->iCntGZones + tgtPage.dlGetNumZones();
	
	if( bDetail ) {				
		out << "Page ID : " << srcPage.dlGetPageID() << endl;
		out << "-------------" << endl;
	}
		
	for( DLPage::DLZonePtrIterator sIter = srcPage.begin();
		 sIter != srcPage.end();
		 sIter++, szZoneTotal++ ) {
		
		string strSType = (*sIter)->dlGetTag("gedi_type");
		this->addType( strSType );
		
		// get zones to be compared
		string strSID = (*sIter)->dlGetZoneID();		
		if( (zone = findZoneByID( tgtPage, strSID )) == NULL ) {
			cout << "Couldn't find mathcing zone( ID : " << strSID << " ) on Ground-truth" << endl;
			continue;
			// exit(0);
		}
		
		string strTType = zone->dlGetTag("gedi_type");
		this->addType( strTType );
		
		// update confusion matrix			
		this->iConfMatrix[this->mapZType[strSType]][this->mapZType[strTType]]++;
		
		// print out the result of each zone
		if( this->bDetail ) {
			
			size_t szSpace = getTypeMaxLength() + 1;
			
			if( strSType == strTType ) {
				
				szMatchCnt++;
				
				out << "[O] "
					<< setw(5) << right << strSID
					<< ","
					<< setw(szSpace) << right << strSType
					<< endl;
			} else {
				
				szUnmatchCnt++;
				
				out << "[X] "
					<< setw(5) << right << strSID
					<< ","
					<< setw(szSpace) << right << strSType
					<< ","
					<< setw(szSpace) << right << strTType
					<< endl;
			}
		}
		
		// modify page for visualization
		if( this->bVisual ) {			
			(*sIter)->dlSetTag("RS_Type", strSType, true);
			(*sIter)->dlSetTag("GT_Type", strTType, true);
			
			if( strSType == strTType )	
				(*sIter)->dlSetTag("gedi_type", "DL_CORRECT", true);
			else
				(*sIter)->dlSetTag("gedi_type", "DL_CORRECT", true);			
		}		
	}
	
	// print out the summary result of page
	if( this->bDetail ) {
		
		double dMatchRate = (double)szMatchCnt / szZoneTotal * 100;
		
		out << "[OVERALL] "
			<< szMatchCnt
			<< "/"
			<< szZoneTotal
			<< ", "
			<< fixed << setprecision(2) << dMatchRate
			<< "%"
			<< endl
			<< endl;
	}
}

void DLPets::compareLPage( DLPage& srcPage, DLPage& tgtPage, ofstream & out ) {
	
	UINT iSMZoneSize = 0;	
	UINT iTMZoneSize = 0;	
	
	// to do : calculate the size of array exactly
	DLMZone * srcMZones = new DLMZone[ srcPage.dlGetNumZones() * 2 ];
	DLMZone * tgtMZones = new DLMZone[ tgtPage.dlGetNumZones() ];	
	
	std::list<Score> lstScore;
	std::list<Score> result;
	
	// getZones( srcPage, srcMZones, iSMZoneSize );
	getZones( tgtPage, tgtMZones, iTMZoneSize );
	
	if( bMerge ) getMZones( srcPage, srcMZones, iSMZoneSize, tgtMZones, iTMZoneSize );
	else 		 getZones( srcPage, srcMZones, iSMZoneSize );
	
	this->iCntRZones = this->iCntRZones + srcPage.dlGetNumZones();
	this->iCntGZones = this->iCntGZones + tgtPage.dlGetNumZones();
	
	// 2-dimension table for matching score
	double** fScoreTable;
	fScoreTable = new double*[iSMZoneSize];
	for( size_t i = 0; i < iSMZoneSize; i++ ) fScoreTable[i] = new double[iTMZoneSize];	
	
	// cout << "sm = " << iSMZoneSize << endl;
	// cout << "tm = " << iTMZoneSize << endl;	
	
	for( size_t row = 0; row < iSMZoneSize; row++ ) {				
		for( size_t col = 0; col < iTMZoneSize; col++ ) {					
			if(srcMZones[row].isOverlap( tgtMZones[col].getRect() )) {				
				fScoreTable[row][col] = srcMZones[row].getMatchScore( tgtMZones[col] );
				// cout << "[" << row << "][" << col << "] = " << fScoreTable[row][col] << endl;
				if( fScoreTable[row][col] >= this->dMatchTh ) {
					Score score = { fScoreTable[row][col], row, col };
					lstScore.push_back( score );					
				}
			} else {				
				fScoreTable[row][col] = 0.0;
			}
		}
	}
	
	lstScore.sort( compare_function );		// sort	
	
	result = getMaxMatch( lstScore, srcMZones, tgtMZones );		
	
	/*
	 * Analyse the matching score table
	 */	
	
	// matching table
	UINT** iMatchTbl;
	iMatchTbl = new UINT*[iSMZoneSize];
	for( size_t i = 0; i < iSMZoneSize; i++ ) iMatchTbl[i] = new UINT[iTMZoneSize];
	
	for( UINT i = 0; i < iSMZoneSize; i++ )
		for( UINT j = 0; j < iTMZoneSize; j++ ) iMatchTbl[i][j] = MISS;			
	
	std::list<Score>::iterator rEIt = result.end();
	for( std::list<Score>::iterator rBIt = result.begin();
	     rBIt != rEIt;	    
	     rBIt++ ) {		
		
		// cout << "[" << rBIt->row << "][" << rBIt->col << "] = " << rBIt->dlScore << endl;
		
		if( srcMZones[rBIt->row].getType() == tgtMZones[rBIt->col].getType() ) 
			iMatchTbl[rBIt->row][rBIt->col] = MATCH;
		else iMatchTbl[rBIt->row][rBIt->col] = DETECT;
	}
	
	updateConfMatrix( iSMZoneSize, iTMZoneSize, srcMZones, tgtMZones, iMatchTbl );	
	
	if( bDetail ) {		
		out << "Page ID : " << srcPage.dlGetPageID() << endl;
		out << "-------------" << endl;					
		printDetail( iSMZoneSize, iTMZoneSize, srcMZones, tgtMZones, fScoreTable, iMatchTbl, out );	
	}
	if( bVisual ) saveVisual( iSMZoneSize, iTMZoneSize, srcMZones, tgtMZones, srcPage, fScoreTable, iMatchTbl );		
	
	
	//  delocation
	if( fScoreTable ) {
		for( size_t i = 0; i < iSMZoneSize; i++ ) delete[] fScoreTable[i];
		delete[] fScoreTable;
	}
	
	if( iMatchTbl ) {
		for( size_t i = 0; i < iSMZoneSize; i++ ) delete[] iMatchTbl[i];
		delete[] iMatchTbl;
	}
	
	delete[] srcMZones;
	delete[] tgtMZones;	
}

void DLPets::updateConfMatrix( UINT rowSize, UINT colSize, DLMZone* srcMZones, DLMZone* tgtMZones, UINT** table) {
	
	bool bMatch;
	bool bDetect;
	int  iPos;
	
	for( size_t i = 0; i < rowSize; i++ ) {
		bMatch = false;
		bDetect = false;
		
		for( size_t j = 0; j < colSize; j++ ) {
			iPos = 0;
			
			if( table[i][j] == MATCH || table[i][j] == DETECT ) {
				if( table[i][j] == MATCH ) bMatch = true;
				else bDetect = true;
				iPos = j;
				break;
			}
		}		
	
		if( bMatch || bDetect ) {
			// cout << this->mapZType[srcMZones[i].getType()] << ", " << this->mapZType[tgtMZones[iPos].getType()] << " ++" << endl;
			this->iConfMatrix[this->mapZType[srcMZones[i].getType()]][this->mapZType[tgtMZones[iPos].getType()]]++;			
			
			//this->iConfMatrix[this->mapZType[strSZones[i]]][this->mapZType[strTZones[iPos]]]++;
		} else {
			this->iConfMatrix[this->mapZType[srcMZones[i].getType()]][0]++;
		}		
	}
	
	bool fDetect;
	for( UINT i = 0; i < colSize; i++ ) {
		fDetect = false;
		for( UINT j = 0; j < rowSize; j++ ) {
			if( table[j][i] == MATCH || table[j][i] == DETECT ) {
				fDetect = true;
				break;
			}
		}
		if( !fDetect ) {			
			this->iConfMatrix[0][this->mapZType[tgtMZones[i].getType()]]++;	
		}
	}	
}

size_t DLPets::getTypeMaxLength() {
	
	size_t max = 0;
	
	for( size_t i = 0; i < this->iCntType; i++ ) {
		if( this->strZType[i].length() > max ) max = this->strZType[i].length();
	}
	
	return max;
}

void DLPets::printDetail( UINT rowSize, UINT colSize, DLMZone* srcMZones,  DLMZone* tgtMZones, double** fScoreTable, UINT** table, ofstream& out) {
	
	int iTemp;
	int iDetCnt = 0;
	int iFalseCnt = 0;
	int iCorrCnt = 0;		
	
	bool bMatch;
	bool bDetect;
	
	size_t szSpace = getTypeMaxLength() + 1;
		
	//DLPage::DLZonePtrIterator sEIter = srcPage.end();
	//for( DLPage::DLZonePtrIterator sIter = srcPage.begin();
	//     sIter != sEIter;
	//     sIter++ ) {
	for( UINT i = 0; i < rowSize; i++ ) {
		
		bMatch = false;
		bDetect = false;			 
		
		//DLPage::DLZonePtrIterator tEIter = tgtPage.end();
		//for( DLPage::DLZonePtrIterator tIter = tgtPage.begin();
		//     tIter != tEIter;
		//     tIter++ ) {
		for( UINT j = 0; j < colSize; j++ ) {				
			
			iTemp = 0;
			
			if( table[i][j] == MATCH || table[i][j] == DETECT ) {
				if( table[i][j] == MATCH ) bMatch = true;
				else bDetect = true;
				iTemp = j;
				break;
			}			
		}
		
		if( bMatch ) {
					
			iCorrCnt++;
			
			out << "[O] "
				<< setw(5) << right << srcMZones[i].getLineID()
				<< ", "
				<< setw(szSpace) << right << srcMZones[i].getType()
				<< ", "
				<< setw(5) << right << tgtMZones[iTemp].getLineID()
				<< ", "
				<< setw(szSpace) << right << tgtMZones[iTemp].getType()
				<< ", " 
				<< fixed << setprecision(2) << fScoreTable[i][iTemp]
				<< "%"
				<< endl;
			
		} else if( bDetect ) {
			
			iDetCnt++;
			
			out << "[-] "
				<< setw(5) << right << srcMZones[i].getLineID()
				<< ", "
				<< setw(szSpace) << right << srcMZones[i].getType()
				<< ", "
				<< setw(5) << right << tgtMZones[iTemp].getLineID()
				<< ", "
				<< setw(szSpace) << right << tgtMZones[iTemp].getType()
				<< ", " 
				<< fixed << setprecision(2) << fScoreTable[i][iTemp]
				<< "%"
				<< endl;
						
		} else {
			
			iFalseCnt++;
			
			out << "[X] "
				<< setw(5) << right << srcMZones[i].getLineID()
				<< ", "
				<< setw(szSpace) << right << srcMZones[i].getType()
				<< endl;
		}
		
		if( srcMZones[i].getZoneCnt() >= 2 ) {				
			out << "    "
			    << setw(5) << right << srcMZones[i].getLineID()
			    << " : "
			    << srcMZones[i].getZoneListStr()
			    << endl;
		}
	}
	
	out << "[OVERALL] "
	    << iCorrCnt		    
	    << "/" 
	    << iDetCnt
	    << "/"
	    << iFalseCnt
	    << "/"
	    << rowSize 
	    << ", "
	    << fixed << setprecision(2) << (double)iCorrCnt/rowSize*100
	    << "%"
	    << endl;
	out << endl;			
}

/*
string DLPets::getRLECode( DLZone * zone ) {
	DLMZone temp(0, "", "", zone);	
	return temp.getRunLengthCode();
}
*/

void DLPets::saveVisual( UINT iSMZoneSize, UINT iTMZoneSize, DLMZone* srcMZones, DLMZone* tgtMZones, DLPage& srcPage, double** fScoreTable, UINT** iMatchTbl ) {

	bool bMatch;
	bool bDetect;
	
	int pos;	
	string score;
	double dMax=0.0;	
	
	for( DLPage::DLZonePtrIterator sIter = srcPage.begin();
	     sIter != srcPage.end();
	     sIter++ ) {		 
	
	//for( UINT i = 0; i < iSMZoneSize; i++ ) {
		
		bMatch = false;
		bDetect = false;
		
		int i = cvStrToInt( (*sIter)->dlGetTag("MZID") );			
		
		//for( DLPage::DLZonePtrIterator tIter = tgtPage.begin();
		//     tIter != tgtPage.end();
		//     tIter++ ) {
		for( size_t j = 0; j < iTMZoneSize; j++ ) {
			
			// int j = cvStrToInt( (*tIter)->dlGetTag("MZID") );
			
			pos = 0;
			
			if( iMatchTbl[i][j] == MATCH || iMatchTbl[i][j] == DETECT ) {
				if( iMatchTbl[i][j] == MATCH ) bMatch = true;
				else bDetect = true;					
				pos = j;
				break;
			}				 
		}
		
		(*sIter)->dlSetTag("RESID", srcMZones[i].getLineID());
		(*sIter)->dlSetTag("RESClass", srcMZones[i].getType());		
		
		if( this->bRunLength && !(*sIter)->dlIsTagSet("RLEIMAGE") )
			(*sIter)->dlSetTag("RLEIMAGE", srcMZones[i].getRunLengthCode(), true );
		
		if( bMatch ) {
			(*sIter)->dlSetTag("GTID", tgtMZones[pos].getLineID());
			(*sIter)->dlSetTag("GTClass", tgtMZones[pos].getType());
			(*sIter)->dlSetTag("gedi_type", "MATCHED", true); // should be changed to gedi_type at the final
			
			score = cvDblToStr( fScoreTable[i][pos] );				
			(*sIter)->dlSetTag("score", score, true );
		} else if( bDetect ) {
			(*sIter)->dlSetTag("GTID", tgtMZones[pos].getLineID());
			(*sIter)->dlSetTag("GTClass", tgtMZones[pos].getType());
			(*sIter)->dlSetTag("gedi_type", "DETECTED", true); // should be changed to gedi_type at the final
			
			score = cvDblToStr( fScoreTable[i][pos] );
			(*sIter)->dlSetTag("score", score, true );				
		} else {
			dMax = 0.0;
			(*sIter)->dlSetTag("gedi_type", "FALSEALRAM", true);
			
			for( size_t k = 0; k < iTMZoneSize; k++ ) {
				if( fScoreTable[i][k] > dMax ) {
					dMax = fScoreTable[i][k];
					score = cvDblToStr( dMax );
				}
			}
			(*sIter)->dlSetTag("score", score, true );				
		}
		
		//if( this->bRunLength && !(*sIter)->dlIsTagSet("RLEIMAGE") )
		//	(*sIter)->dlSetTag("RLEIMAGE", getRLECode( *sIter ), true );
		
		
		// (*sIter)->dlDeleteTag("MZID");
	}
		
	bool fDetect;
	for( size_t col = 0; col < iTMZoneSize; col++ ) {
		fDetect = false;
		for( size_t row = 0; row < iSMZoneSize; row++ )
			if( iMatchTbl[row][col] == MATCH || iMatchTbl[row][col] == DETECT ) fDetect = true;
		
		if( !fDetect ) {

			DLZone newZone;
			
			char buf[10];
			sprintf(buf, "MISS%d", ++iMaxZoneNum);
			
			DLRectShape rect = tgtMZones[col].getRect();
			
			newZone.dlSetZoneID( string(buf) );
			newZone.dlSetZoneOrigin( rect.dlGetX(), rect.dlGetY() );
			newZone.dlSetZoneWidth( rect.dlGetWidth() );
			newZone.dlSetZoneHeight( rect.dlGetHeight() );
			newZone.dlSetTag("gedi_type", "MISSED", true );
			newZone.dlSetTag("GTID", tgtMZones[col].getLineID() );
			newZone.dlSetTag("GTClass", tgtMZones[col].getType() );
			
			dMax = 0.0;
			
			for( size_t k = 0; k < iSMZoneSize; k++ ) {
				if( fScoreTable[k][col] > dMax ) {
					dMax = fScoreTable[k][col];
					score = cvDblToStr( dMax );
				}
			}
			newZone.dlSetTag("score", score, true);
			if( this->bRunLength ) newZone.dlSetTag("RLEIMAGE", tgtMZones[col].getRunLengthCode(), true );
			
			// todo
			// Zone boundary check. there is error some time.
			
			try {	
				srcPage.dlAppendZoneCopy( newZone );
			} catch(...) {
				cout << "Append zone is failed" << endl;
			}				

		}
	}	
}

std::list<Score> DLPets::getMaxMatch( std::list<Score>& lstScore, DLMZone* srcMZones, DLMZone* tgtMZones ) {
	
	std::list<Score> temp;
	std::list<Score> result;
	
	int iMaxMCnt = 0;
	int iMaxDCnt = 0;
	int iMaxMDCnt = 0;
	
	int iMatchCnt = 0;
	// int iMatchDCnt = 0;
	int iDetectCnt = 0;	
	
	for( std::list<Score>::iterator bIt = lstScore.begin();
		 bIt != lstScore.end();
		 bIt++ ) {
		
		for(std::list<Score>::iterator it = bIt;
		    it != lstScore.end();
		    it++ ) temp.push_back(*it);		
		
		for( std::list<Score>::iterator tIt = temp.begin();
		     tIt != temp.end();
		     tIt++ ) {
			int iRow = tIt->row;
			int iCol = tIt->col;			
			
			for( std::list<Score>::iterator tmpIt = tIt;
			  tmpIt != temp.end();
			  tmpIt++ ) {
				
				if( tmpIt == tIt ) continue;
				
				if( tmpIt->row == iRow || tmpIt->col == iCol ) {					
					tmpIt = temp.erase( tmpIt );
					tmpIt--;
					// cout << "Deleted" << endl;
				}				
			}			
		}
		// cout << "Here" << endl;
		iMatchCnt = 0;
		iDetectCnt = 0;
		
		for( std::list<Score>::iterator tIt = temp.begin();
		     tIt != temp.end();
		     tIt++ ) {
			
			if( srcMZones[tIt->row].getType() == tgtMZones[tIt->col].getType() )
				iMatchCnt++;
			else iDetectCnt++;
			
			// if( strSZones[tIt->row] == strTZones[tIt->col] ) iMatchCnt++;
		}
		// cout << "iMatchCnt = " << iMatchCnt << endl;
		
		if( iMatchCnt > iMaxMCnt) {			
			result = temp;
			iMaxMCnt = iMatchCnt;
			iMaxMDCnt = iDetectCnt;
		} else if ( iMatchCnt == iMaxMCnt ) {
			if( iDetectCnt > iMaxMDCnt ) {
				result = temp;
				iMaxMDCnt = iDetectCnt; 
			}
		} else {
			if( iMaxMCnt == 0 && iDetectCnt > iMaxDCnt ) {
				result = temp;
				iMaxDCnt = iDetectCnt;
			}
		}		
		temp.clear();
	}
	
	return result;
}

void DLPets::printIntro( ofstream & fout ) {
	
	time_t rowtime;
	struct tm* timeinfo;
	
	time( &rowtime );
	timeinfo = localtime( &rowtime );
	
	fout << "Result of Performance Evaluation ToolS for Zone Segmentation and Classification(PETS)" << endl;	
	fout << "Generated from " << asctime( timeinfo ) << endl;
}

void DLPets::printEnd( ofstream & fout ) {
	
	time_t rowtime;
	struct tm* timeinfo;
	
	time( &rowtime );
	timeinfo = localtime( &rowtime );
	
	fout << endl;
	fout << "Evaluation process is finished at " << asctime( timeinfo ) << endl;	
}

UINT DLPets::getRCntZone( string type ) {
	UINT count = 0;	
	for( UINT i = 0; i < this->iCntType; i++ )
		count = count + this->iConfMatrix[this->mapZType[type]][i];
	
	return count;
}
UINT DLPets::getGCntZone( string type ) {
	UINT count = 0;
	for( UINT i = 0; i < this->iCntType; i++ )
		count = count + this->iConfMatrix[i][this->mapZType[type]];
	
	return count;
}

UINT DLPets::getAccZone( string type ) {
	return this->iConfMatrix[this->mapZType[type]][this->mapZType[type]];
}

UINT DLPets::getFalseZone( string type ) {
	return getRCntZone(type)-getAccZone(type);
}

UINT DLPets::getMissZone( string type ) {
	return getGCntZone(type)-getAccZone(type);
}

double DLPets::getAccZoneRate( string type ) {
	if( getRCntZone(type) != 0 )
		return (double)getAccZone(type) / getRCntZone(type) * 100;
	else return 0.0;
}
double DLPets::getMissZoneRate( string type ) {
	if( getGCntZone(type) != 0 )
		return (double)getMissZone(type) / getGCntZone(type) * 100;
	else return 0.0;
}
double DLPets::getFalseZoneRate( string type ) {
	if( getRCntZone(type) != 0 )
		return (double)getFalseZone(type) / getRCntZone(type) * 100;
	else return 0.0;
}

double DLPets::getPreciZone( string type ) {
	if( getRCntZone(type) != 0 )
		return (double)getAccZone(type)/getRCntZone(type) * 100;
	else return 0.0;
}

double DLPets::getRecallZone( string type ) {
	if( getGCntZone(type) != 0 ) 
		return (double)getAccZone(type)/getGCntZone(type) * 100;
	else return 0.0;
}

double DLPets::getFScore( double precision, double recall ) {
	if( (precision + recall) > 0 ) return (2.0 * precision * recall) / ( precision + recall );
	else return 0.0;
}

size_t DLPets::calSpace() {
	
	UINT big = 0;
	
	for( UINT i = 0; i < this->iCntType; i++ ) {
		for( UINT j = 0; j < this->iCntType; j++ ) {
			if( this->iConfMatrix[i][j] > big ) {
				big = this->iConfMatrix[i][j];
			}
		}
	}
	if( big < 10 ) return 2;
	else if( big < 100 ) return 3;
	else if( big < 1000 ) return 4;
	else if( big < 10000 ) return 5;
	else if( big < 100000 ) return 6;
	else if( big < 1000000 ) return 7;
	else return 10;
}

void DLPets::summary( ofstream & out ) {
	
	out << "==================" << endl;
	out << "Summary of Results" << endl;
	out << "==================" << endl;
	out << fixed << endl; 
	
	UINT start;
	if( bZC ) start = 0;
	else      start = 1;
	
	out << "- Number of Files : " << this->iFileCnt << endl;
	//out << "- Number of Zones in Result : " << this->iCntRZones << endl;
	//if( !bZC ) out << "- Number of Zones in Ground truth : " << this->iCntGZones << endl;
	
	double dAccuracy;
	// UINT iCorrect = 0;	
	// for( UINT i = start; i < this->iCntType; i++ ) iCorrect = iCorrect + this->iConfMatrix[i][i];
	double df1total = 0.0;
	for( UINT i = start; i < this->iCntType; i++ ) 
		df1total = df1total + getFScore( getPreciZone(this->strZType[i] ), getRecallZone(this->strZType[i]) ); 
	
	dAccuracy = df1total / this->iCntType;
	if( bZC ) out << "- Accuracy of Zone Matching(average) : " << setprecision(2) << dAccuracy << "%" << endl; 
	else out << "- Accuracy of Zone Detecting(average) : " << setprecision(2) << dAccuracy << "%" << endl;
	
	out << endl;	
	
	out << "01. Information on Zones" << endl;
	out << "========================" << endl;
	out << endl;
	out << setw(8) << left << "Label"
	    << setw(20) << left << "Class of Zone" 
	    << setw(20) << right << "Num. in Result"
	    << setw(20) << right << "Num. in G-Truth"
	    << setw(15) << right << "Accuracy(F1)"	    
	    << endl;
	
	for( size_t i = 0; i < 83; i++ ) out << "-";
	out << endl;
	
	for( UINT i = start; i < this->iCntType; i++ ) {		
		out << setw(8) << left << this->mapZType[this->strZType[i]]
		    << setw(20) << left << this->strZType[i]
		    << setw(20) << right << getRCntZone( this->strZType[i] )
		    << setw(20) << right << getGCntZone( this->strZType[i] )
		    // << setw(14) << setprecision(2) << right << getAccZoneRate( this->strZType[i] )
		    // change to F1 score for the accuracy
		    << setw(14) << setprecision(2) << right << getFScore( getPreciZone(this->strZType[i] ), getRecallZone(this->strZType[i]) )
		    << setw(1) << "%"
		    << endl;		
	}	
	out << endl;
	
	out << "02. Confusion Matrix" << endl;
	out << "====================" << endl;	
	out << endl;
	
	size_t space = calSpace();
	size_t Tspace = space + 8;
	
	out << setw(10) << right << "Result\\GT";	
	
	for( UINT i = 0; i < this->iCntType; i++ ) {
		if( i == 0 ) {
			if( !bZC ) out << setw(Tspace-1) << right << "unmatch";
			else 	   out << setw(Tspace-1) << right << this->mapZType["unmatch"];
		} else out << setw(Tspace) << right << this->mapZType[this->strZType[i]];
	}
	out << endl;
	
	out << "---------";
	for( UINT i = 0; i < this->iCntType; i++ )
		for( int j = 0; j < Tspace; j++ ) out << "-";		
	out << endl;
	
	double dvalue;
	string asteric;
	for( UINT i = 0; i < this->iCntType; i++ ) {
		
		if( i == 0 ) {
			if( !bZC ) out << setw(10) << right << "unmatch";
			else	   out << setw(10) << right << this->mapZType[this->strZType[i]];
		} else out << setw(10) << right << this->mapZType[this->strZType[i]];
		
		for( UINT j = 0; j < this->iCntType; j++ ) {
			
			if( !bZC ) {
				if( i == 0 ) {
					if( getGCntZone(this->strZType[j]) > 0 ) dvalue = (double)this->iConfMatrix[i][j] / getGCntZone(this->strZType[j]) * 100;
					else dvalue = 0.0;
				} else {
					if( getRCntZone(this->strZType[i]) > 0 ) dvalue = (double)this->iConfMatrix[i][j] / getRCntZone(this->strZType[i]) * 100;
					else dvalue = 0.0;
				}
			} else {
				if( getRCntZone(this->strZType[i]) > 0 ) dvalue = (double)this->iConfMatrix[i][j] / getRCntZone(this->strZType[i]) * 100;
				else dvalue = 0.0;
			}
			
			if( i == j ) asteric = "*";
			else asteric = " ";

			out << fixed
			    << setw(space) << right << this->iConfMatrix[i][j]
			    << setw(1) << "(";
			
			if( dvalue < 100.0 ) out << setw(4) << right << setprecision(1) << dvalue;
			else out << setw(4) << right << setprecision(0) << dvalue;
			
			out << setw(1) << "%"
			    << setw(1) << ")"
			    << setw(1) << asteric;			    			
		}
		out << endl;
	}
	out << endl;
	
	out << "03. Result Table" << endl;
	out << "================" << endl;
	out << endl;
	
	out << setw(5) << right << "Label"	     
	    << setw(8) << right << "GT"
	    << setw(10) << right << "Result" 
	    << setw(10) << right << "Correct"
	    << setw(10) << right << "Precsion"	    
	    << setw(10) << right << "Recall"	
		<< setw(10) << right << "F-Score";
	if( !bZC ) {
	    out	<< setw(10) << right << "Missing"
	    	<< setw(12) << right << "FalseAlarm";	    	
	}
	out << endl;

	UINT iDLength;
	if( bZC ) iDLength = 53;
	else iDLength = 85;
	
	for( UINT i = 0; i < iDLength; i++ ) out << "-";
	out << endl;	

	for( UINT i = 1; i < this->iCntType; i++ ) {		
		
		out << setw(5) << right << this->mapZType[this->strZType[i]]
		    << setw(8) << right << getGCntZone(this->strZType[i])
		    << setw(10) << right << getRCntZone(this->strZType[i])
		    << setw(10) << right << getAccZone(this->strZType[i])
		    << fixed
		    << setw(9) << right << setprecision(2) << getPreciZone(this->strZType[i]) << setw(1) << "%"
		    << setw(9) << right << setprecision(2) << getRecallZone(this->strZType[i]) << setw(1) << "%"				
		    << setw(9) << right << setprecision(2) << getFScore( getPreciZone(this->strZType[i] ), getRecallZone(this->strZType[i]) ) << setw(1) << "%";
		
		if( !bZC ) {
		    out << setw(9) << right << setprecision(2) << getMissZoneRate(this->strZType[i]) << setw(1) << "%"
		    	<< setw(11) << right << setprecision(2) << getFalseZoneRate(this->strZType[i]) << setw(1) << "%";
		}
		    out << endl;		
	}		
}

#if defined (WIN32) || defined(_WINDOWS) || defined(__WIN32__) || defined(_Windows)
	const string sep = "\\";
#else
	const string sep = "/";
#endif

void DLPets::doEval() {
	
	string src;
	string target;
	string img;
	string save;	
	
	if( !bZC) addType( string("unmatch") );
	
	ofstream fout( this->rPath.string().c_str() );
	this->printIntro( fout );	
	
	fout << "Command Option : " << this->strOptions << endl;
	fout << endl;
	
	if( bDetail ) {
		fout << "==========================" << endl;
		fout << "Result of Individual File" << endl;
		fout << "==========================" << endl;
		fout << endl;
		
		if( bZC ) {
			fout << "[O] : Matched, [X] : Unmatched" << endl;
		} else {
			fout << "[O] : Matched, [-] : Detected, [X] : False Alarm" << endl;
		}
		
		fout << endl;
	}		
	
	if(this->bMultiToMulti) {
		
		std::list<string> slist;
		
		fs::directory_iterator end_itr;
		for( fs::directory_iterator itr( this->oPath );
		     itr != end_itr;
		     itr++ ) {
			
			if( fs::is_directory( itr->status() ) ) continue;
			if( !strstr( itr->leaf().c_str(), ".xml" ) ) continue;
			if( !strncmp( itr->leaf().c_str(), "GEDIConfig", 10) ) continue;
			
			slist.push_back( itr->leaf() );
			
		}		
		
		slist.sort();
		string fName;
		
		for( std::list<string>::iterator iter = slist.begin();
			 iter != slist.end();
			 iter++ ) {
			
			fName = *iter;
			
			src = this->oPath.string() + sep + fName;
			// src = fs::path( this->oPath sep fName.c_str() );
			
			target = this->gPath.string() + sep + fName;
			// target = fs::path( this->gPath sep fName.c_str() );			
			if( !fs::exists( fs::path( target.c_str() ) ) ) {
				cout << target << " is not found!!!" << endl;
				continue;
			}
			
			save = this->vPath.string() + sep + fName;
			// save = fs::path( this->vPath sep fName.c_str() );
			
			string savename = save;			
			savename.insert( savename.size()-3, "evl.");	
			compareDoc(src, target, savename, fout);
		}
		
	}else if(this->bMultiToSingle) {		
		
	}else if(this->bSingleToMulti) {
		
	}else if(this->bSingleToSingle) {
		save = this->vPath.string() + sep + this->oPath.string();
		// save = fs::path( this->vPath / this->oPath );
		string savename = save;			
		savename.insert( savename.size()-3, "evl.");
		compareDoc( this->oPath.string(), this->gPath.string(), savename, fout );
	}
	
	summary( fout );	
	this->printEnd( fout );
	
	fout.close();	
}