#ifndef DLPETS_H_
#define DLPETS_H_

#include <fstream>
#include <string>
#include <list>

// from Doclib
#include "DLRectShape.h"
#include "DLDocument.h"

#include "DLPolygon.h"
#include "DLMZone.h"

#include "boost/filesystem/fstream.hpp"
#include "boost/filesystem/path.hpp"
namespace fs = boost::filesystem;

typedef unsigned int UINT;

#define MATCH 	2
#define DETECT 	1
#define MISS 	0

#define DEFAULT_THRESHOLD 80.0

struct Score {
	double dlScore;
	int row;
	int col;
};

class DLPets {
	// member variable
private:	
	
	bool bMultiToMulti;
	bool bMultiToSingle;
	bool bSingleToMulti;
	bool bSingleToSingle;
	
protected:
	
	int iFileCnt;
	
	string strOptions;
	
	int iMaxZoneNum;
	
	double dMatchTh;
	double mergeTH;
	
	bool bThreshold;
	
	string * strZType;	
	map<string, int> mapZType;
	
	size_t hello;
	
	
	UINT iCntType;
	// Confusion matrix
	UINT** iConfMatrix;
	
	// Count of Groun Truth Zones
	UINT iCntGZones;
	// Count of Result Zones
	UINT iCntRZones;
	
	// Count of Types of Zones
	UINT iCntSRectZones;
	UINT iCntSPolyZones;
	UINT iCntSOriZones;
	
	UINT iCntTRectZones;
	UINT iCntTPolyZones;
	UINT iCntTOriZones;
	
	// FalseAlram Array
	UINT* iFalseAlram;
	// Missing Array
	UINT* iMissing;
	
	// Name of file to flush the result of evaluation out
	char* strRFileName;
	
	bool bSingleInput;
	// char* strGFileName;
	// char* strOFileName;
	
	bool bDetail;
	bool bVisual;
	bool bSegOnly;
	bool bZC;
	
	bool bAZ;
	bool bNAZ;
	
	bool bLineID;
	bool bMerge;
	bool bRunLength;
	
	string strRRoot;
	string strGRoot;
	string strImgRoot;
	string strSaveRoot;
	string strOPath;
	
	list<string> fileList;
	
	vector<string> accZoneList;
	vector<string> nAccZoneList;
	
	fs::path gPath;		// ground truth path
	fs::path oPath;		// program output path
	
	fs::path rPath;		// evaluation output path
	fs::path imgPath;	// image path
	fs::path vPath;		// visualization path
	
	vector<string> mergeList;
	
public:
	
	// member funcion
private:
	
	size_t getTypeMaxLength();	
	
	string cvIntToStr( int num );
	int cvStrToInt( string str );
	string cvDblToStr( double num );
	
	// void setIDType( DLPage& page, string* id, string* type );
	
	bool checkAccZone( string type );
	bool checkNotAccZone( string type );
	
	bool isMergeList( string type );
	
	size_t calSpace();	
	bool exist( string type );
	void addType( string type );
	void initialize();
	void parseOptions(int argc, char** argv);
	void printIntro( std::ofstream & fout );
	void printEnd( std::ofstream & fout );
	
	bool isOverlapped( const DLRectShape src, const DLRectShape target );
	
	int transPosX( int x, int y, double angle );
	int transPosY( int x, int y, double angle );
	UINT countPixels( DLZone& zone, DLImage & img );
	UINT countOverlappedPixels( DLZone& srcZone, DLZone& tgtZone, DLImage & pageImg );
	DLRectShape interSection( DLRectShape* src, DLRectShape* tgt );
	
	UINT getGCntZone( string type );
	UINT getRCntZone( string type );
	UINT getAccZone( string type );
	UINT getMissZone( string type );
	UINT getFalseZone( string type );
	
	double getAccZoneRate( string type );
	double getMissZoneRate( string type );
	double getFalseZoneRate( string type );
	
	double getPreciZone( string type );
	double getRecallZone( string type );
	
	double getFScore( double precision, double recall );
	
	DLZone* findZoneByID( DLPage& page, string id );
	
protected:
	
	list<Score> getMaxMatch( list<Score>& lstScore, DLMZone* srcMZones, DLMZone* tgtMZones );
	void updateConfMatrix( UINT rowSize, UINT colSize, DLMZone* srcMZones, DLMZone* tgtMZones, UINT** table);
	void printDetail( UINT rowSize, UINT colSize, DLMZone* srcMZones,  DLMZone* tgtMZones, double** fScoreTable, UINT** table, std::ofstream& out);
	
	void saveVisual( UINT iSMZones, UINT iTMZones, DLMZone* srcMZones, DLMZone* tgtMZones, DLPage& srcPage, double** fScoreTable, UINT** iMatchTbl );
	
	DLPolygon getPolygon( DLZone & zone );
	DLPolygon getPolyFromOri( DLZone & zone );
	
	//string getRLECode( DLZone * zone );
	
	DLRectShape* getOuterRectPoly( DLZone & zone );
	DLRectShape* getOuterRectOri( DLZone& zone );
	DLRectShape* getRect( DLZone& zone );
	
	
	// void addZone( DLMZone * zones, UINT& size, int id, string lid, string type, DLZone * zone);
	
	void getZones( DLPage& page, DLMZone * zones, UINT& sizeM );
	void getMZones( DLPage& page, DLMZone * zones, UINT& size, DLMZone * tZones, UINT tSize );

	map<string,DLRectShape> getRectMap( DLPage& page );
	void compareDoc( const string& src, const string& target, const string& save, std::ofstream & fout );
	void compareCPage( DLPage& srcPage, DLPage& tgtPage, ofstream & out );
	void compareLPage( DLPage& srcPage, DLPage& tgtPage, ofstream & out );
	
	double compareZone(DLZone& src, DLZone& target, DLImage & pageImg );
		
	void summary( ofstream & out );
	
public:
	// Constructor
	DLPets();
	DLPets(int argc, char** argv);
	
	~DLPets();
		
	void doEval();	
};

#endif /*DLPETS_H_*/
