// #include <cstring>
#include <list>
#include <exception>

#include "DLPolygon.h"

// from Doclib
#include "DLRectShape.h"

using namespace std;

DLPolygon::DLPolygon() {
	npoints = 0;	
	this->xpoints = NULL;
	this->ypoints = NULL;
	bounds = DLRectShape();
}

DLPolygon& DLPolygon::operator=( const DLPolygon & poly ) {

	this->npoints = poly.getSize();
	
	if( this->xpoints ) delete[] this->xpoints;
	if( this->ypoints ) delete[] this->ypoints;
	
	this->xpoints = new int[this->npoints];
	this->ypoints = new int[this->npoints];

	memcpy( this->xpoints, poly.getXpoints(), sizeof(int)*this->npoints);
	memcpy( this->ypoints, poly.getYpoints(), sizeof(int)*this->npoints);

	this->bounds = poly.getBounds();

	return *this;
}

DLPolygon::DLPolygon( const DLPolygon & poly ) {
	this->npoints = poly.getSize();
		
	this->xpoints = new int[this->npoints];
	this->ypoints = new int[this->npoints];
	
	memcpy( this->xpoints, poly.getXpoints(), sizeof(int)*this->npoints);
	memcpy( this->ypoints, poly.getYpoints(), sizeof(int)*this->npoints);

	this->bounds = poly.getBounds();
}

DLPolygon::DLPolygon( int * xpoints, int* ypoints, int npoints ) {
	
	if( npoints < MIN_LENGTH ) {
		throw "The number of Polygon point should be more than 3";
	}

	this->npoints = npoints;

	this->xpoints = new int[this->npoints];
	this->ypoints = new int[this->npoints];	

	memcpy( this->xpoints, xpoints, sizeof(int)*this->npoints );
	memcpy( this->ypoints, ypoints, sizeof(int)*this->npoints );

	calculateBounds();
}

DLPolygon::~DLPolygon() {
	if( this->xpoints ) delete[] this->xpoints;
	if( this->ypoints ) delete[] this->ypoints;
}

int DLPolygon::min( int x, int y ) {
	return x <= y ? x : y;
}

int DLPolygon::max( int x, int y ) {
	return x >= y ? x : y;
}

bool DLPolygon::contains( int x, int y) {
	return contains((double) x, (double) y);
}

bool DLPolygon::boundCheck( double x, double y ) {

	int X = bounds.dlGetX();
	int Y = bounds.dlGetY();
	int W = bounds.dlGetWidth();
	int H = bounds.dlGetHeight();

	if( x >= X && x <= (X+W) && y >= Y && y <= (Y+H) ) return true;
	else return false;
}

bool DLPolygon::contains(double x, double y) {
    if (npoints <= 2 || !boundCheck(x, y)) {
	    return false;
	}
	int hits = 0;

	int lastx = xpoints[npoints - 1];
	int lasty = ypoints[npoints - 1];
	
	int curx, cury;

	// Walk the edges of the polygon
	for (int i = 0; i < npoints; lastx = curx, lasty = cury, i++) {
	
		curx = xpoints[i];
		cury = ypoints[i];

	    if (cury == lasty) {
	    	continue;
	    }

	    int leftx;
	    if (curx < lastx) {
			if (x >= lastx) {
			    continue;
			}
			leftx = curx;
	    } else {
			if (x >= curx) {
			    continue;
			}
			leftx = lastx;
	    }

	    double test1, test2;
	    if (cury < lasty) {
			if (y < cury || y >= lasty) {
			    continue;
			}

			if (x < leftx) {
			    hits++;
			    continue;
			}
			test1 = x - curx;
			test2 = y - cury;
	    } else {
			if (y < lasty || y >= cury) {
			    continue;
			}
			if (x < leftx) {
			    hits++;
			    continue;
			}
			test1 = x - lastx;
			test2 = y - lasty;
	    }

	    if (test1 < (test2 / (lasty - cury) * (lastx - curx))) {
	    	hits++;
	    }
	}
	return ((hits & 1) != 0);
}

void DLPolygon::calculateBounds(){

	int boundsMinX = 65532;
	int boundsMinY = 65532;
	int boundsMaxX = -65531;
	int boundsMaxY = -65531;

	for (int i = 0; i < this->npoints; i++) {
	    int x = this->xpoints[i];
	    boundsMinX = min(boundsMinX, x);
	    boundsMaxX = max(boundsMaxX, x);
	    int y = this->ypoints[i];
	    boundsMinY = min(boundsMinY, y);
	    boundsMaxY = max(boundsMaxY, y);	   
	}	

	this->bounds = DLRectShape(boundsMinX,	boundsMinY,
			                   boundsMaxX - boundsMinX,
			                   boundsMaxY - boundsMinY);
}

DLRectShape DLPolygon::getBounds() const{
	return this->bounds;
}

int DLPolygon::getSize() const {
	return this->npoints;
}

int * DLPolygon::getXpoints() const {
	return this->xpoints;
}

int * DLPolygon::getYpoints() const {
	return this->ypoints;
}
