#include "DLxml.h"

// from Doclib
#include "DLPage.h"
#include "DLZone.h"
#include "DLDocument.h"
#include "DLConst.h"
#include "DLTagList.h"
#include "DLException.h"
#include "DLImageFactory.h"
#include "DLBitsPerPixelConverter.h"

#include <map>

using namespace std;

DLxml::DLxml( DLDocument inputDocument, string strFilename )
{
	// cout << "Converting to XML representation" << endl;

	map <string, string> docDynamicAttrs = inputDocument.dlGetAllTags().tagMap;
	map <string, string>::const_iterator docAttrIter;

	xmlDocTree.InsertSiblingElem("DL_DOCUMENT");
	xmlDocTree.AddDynamicAttrib(docDynamicAttrs);
	string docID = inputDocument.dlGetDocumentID();
	if(docID != "") {
		xmlDocTree.AddAttrib("documentID", docID);
	}
	if((docAttrIter = docDynamicAttrs.find("src")) == docDynamicAttrs.end())
	{
		xmlDocTree.AddAttrib("src", strFilename);
	}

	DLDocument::DLPagePtrIterator page_Iter;
	int pageCnt = 0;

	for ( page_Iter = inputDocument.begin(); page_Iter != inputDocument.end(); page_Iter++ )
	{
		// insert page
		DLPage curPage = **(page_Iter);
		char buffer[10];
		
		if( 0 == pageCnt )
		{
			xmlDocTree.InsertChildElem( "DL_PAGE", 2);
			xmlDocTree.Go2ChildLevel();
		}
		else
		{
			xmlDocTree.InsertSiblingElem( "DL_PAGE" );
		}

		map <string, string> pageDynamicAttrs = curPage.dlGetAllTags().tagMap;
		map <string, string>::const_iterator pageAttrIter;

		++pageCnt;
    	xmlDocTree.AddDynamicAttrib( pageDynamicAttrs );
		sprintf(buffer, "%d", curPage.dlGetPageHeight());
		xmlDocTree.AddAttrib( "height", static_cast<string>(buffer) );
		sprintf(buffer, "%d", curPage.dlGetPageWidth());
		xmlDocTree.AddAttrib( "width", static_cast<string>(buffer) );
		string pageID = curPage.dlGetPageID();
		
		if(pageID != "") {
			xmlDocTree.AddAttrib( "pageID", pageID);
		}
		if((pageAttrIter = pageDynamicAttrs.find("src")) == pageDynamicAttrs.end())
		{
			xmlDocTree.AddAttrib("src", strFilename);
		}

		DLPage::DLZonePtrIterator Zones_Iter;
		int zoneCnt = 0;

		for ( Zones_Iter = curPage.begin(); Zones_Iter != curPage.end(); Zones_Iter++ )
		{
			// insert page
			DLZone* curZone = *(Zones_Iter);
			
			if( 0 == zoneCnt )
			{
				xmlDocTree.InsertChildElem( "DL_ZONE", 3 );
				xmlDocTree.Go2ChildLevel();				
			}
			else
			{
				xmlDocTree.InsertSiblingElem( "DL_ZONE");
			}

			++zoneCnt;
		
			int startX = curZone->dlGetZoneOrigin().x;
			int startY = curZone->dlGetZoneOrigin().y;
			int width = curZone->dlGetZoneWidth();
			int height = curZone->dlGetZoneHeight();
			
			xmlDocTree.AddDynamicAttrib( curZone->dlGetAllTags().tagMap );			
			xmlDocTree.AddAttrib( "height", height );
			xmlDocTree.AddAttrib( "width", width );			
			xmlDocTree.AddAttrib( "row", startY );
			xmlDocTree.AddAttrib( "col", startX );
			
			string zoneID = curZone->dlGetZoneID();
			if(zoneID != "") {
				// xmlDocTree.AddAttrib( "zoneID", zoneID );
				xmlDocTree.AddAttrib( "id", zoneID );
			}

			if( curZone->dlHasChildZones() )
				traverseSubtree( curZone, 0, xmlDocTree );
		}

		xmlDocTree.Go2ParentLevel();
	}

	xmlDocTree.Go2ParentLevel();
}

DLxml::DLxml()
{
	// cout << "Creating new XML file" << endl;
	
	/* Populate static attributes of DL_DOCUMENT */
	staticDocumentAttributes.push_back("documentID");

	/* Populate static attributes of DL_PAGE */
	staticPageAttributes.push_back("pageID");
	staticPageAttributes.push_back("width");
	staticPageAttributes.push_back("height");

	/* Populate static attributes of DL_ZONE */
	staticZoneAttributes.push_back("ZoneId");
	// staticZoneAttributes.push_back("gedi_type");
	staticZoneAttributes.push_back("id");
	staticZoneAttributes.push_back("col");
	staticZoneAttributes.push_back("row");
	staticZoneAttributes.push_back("width");
	staticZoneAttributes.push_back("height");	
	
}

bool DLxml::dlSaveToXmlFile( string filename )
{
	return xmlDocTree.SaveXml( filename.c_str() );
}

DLxml::~DLxml(void)
{
}

DLDocument DLxml::dlReadFromXmlFile( string filePath, string imgPath )
{	
	
#if defined (WIN32) || defined(_WINDOWS) || defined(__WIN32__) || defined(_Windows)
	const string sep = "\\";
#else
	const string sep = "/";
#endif		
	
	if( xmlDocTree.LoadXml( filePath ) == false ) {
		cout << "Load XML Error" << endl;
		exit(0);
	}
	
	int iSeq = 0;
	
	string strdoc = xmlDocTree.RetrieveSpecifiedElem( "DL_DOCUMENT" );	
	string docID = "";
	docID = xmlDocTree.GetAttrib( "documentID" );	
	
	xmlDocTree.setXML( strdoc );

	DLDocument outputDocument;
	outputDocument.dlSetDocumentID(docID);
	
	map <string, string> docDynamicAtt = xmlDocTree.GetDynamicAttrib( 1 );
	map <string, string>::const_iterator docAttr_Iter;
	string docAttrName;
	vector <string>::const_iterator docTag_Iter;
	bool bStaticDocAttr;
	for(docAttr_Iter = docDynamicAtt.begin(); docAttr_Iter != docDynamicAtt.end(); docAttr_Iter++)
	{
		bStaticDocAttr = false;
		docAttrName = (*docAttr_Iter).first;
		if(!staticDocumentAttributes.empty()) 
		{
			for(docTag_Iter = staticDocumentAttributes.begin(); docTag_Iter != staticDocumentAttributes.end(); docTag_Iter++)
			{
				if(docAttrName == *docTag_Iter) {
					bStaticDocAttr = true;
					break;
				}
			}
		}
		if(!bStaticDocAttr) {
			outputDocument.dlSetTag( (*docAttr_Iter).first, (*docAttr_Iter).second, true );
		}
	}
	
	string strSrcFile;
	strSrcFile.append( imgPath );
	strSrcFile.append( sep );		
	
	// Todo
	// modify later...		
	strSrcFile.append( xmlDocTree.GetAttrib( "src" ) );

	// traverse xml file to generate DLDocument structure.
	if( xmlDocTree.bHasChild() )
	{
		// append page
		xmlDocTree.Go2ChildLevel();
	}
	else
		return outputDocument; // no page

	do{				
		
		string pageID = "";
		pageID = xmlDocTree.GetAttrib( "pageID" );		
		
// modified	to load the page image according to the page number
		// DLImage dlImg( strSrcFile.c_str() );		
		
		DLImage dlImg = DLImageFactory::dlGetInstance()->dlLoadImageAtPage( strSrcFile.c_str(), ++iSeq );		
				
		if( dlImg.dlGetDepth() != DLImageHead::DL_BITDEPTH_1 
			&& dlImg.dlGetDepth() != DLImageHead::DL_BITDEPTH_8 ) { 
			// dlImg = DLBitsPerPixelConverter::dlConvertImage( dlImg, DLBitsPerPixelConverter::DL_2GRAY );
				dlImg = DLBitsPerPixelConverter::dlDownscaleColor2Gray_global( dlImg );
		}		
				
		DLPage *dlPage = new DLPage( dlImg, pageID.c_str() );
		dlPage->dlSetPageID( pageID );

		map <string, string> pageDynamicAtt = xmlDocTree.GetDynamicAttrib( 1 );
		map <string, string>::const_iterator pageAttr_Iter;
		string pageAttrName;
		vector <string>::const_iterator pageTag_Iter;
		bool bStaticPageAttr;
		for(pageAttr_Iter = pageDynamicAtt.begin(); pageAttr_Iter != pageDynamicAtt.end(); pageAttr_Iter++)
		{
			bStaticPageAttr = false;
			pageAttrName = (*pageAttr_Iter).first;
			if(!staticPageAttributes.empty()) 
			{
				for(pageTag_Iter = staticPageAttributes.begin(); pageTag_Iter != staticPageAttributes.end(); pageTag_Iter++)
				{
					if(pageAttrName == *pageTag_Iter) {
						bStaticPageAttr = true;
						break;
					}
				}
			}
			if(!bStaticPageAttr) {
				dlPage->dlSetTag( (*pageAttr_Iter).first, (*pageAttr_Iter).second, true );
			}
		}

		// has zones, travel
		if( xmlDocTree.bHasChild() ) 
		{
			bool bChild = xmlDocTree.Go2ChildLevel();

			do{
				DLZone *dlRootZone = new DLZone();
				traverseXML( xmlDocTree, dlRootZone );
				
				if( dlRootZone->dlIsTagSet("polygon") )
					if( !validatePolygon( dlRootZone, dlPage->dlGetPageWidth(), dlPage->dlGetPageHeight() ) ) {
						// cout << "validate" << endl;
						continue;
					}
				
				try {
					dlPage->dlAppendZone( dlRootZone );					
				} catch( DLException e ) {
					cout << "Appending Zone(ID : " << dlRootZone->dlGetZoneID() << ") to the page is canceled due to the size check routine" << endl;				
				}
				
				// delete dlRootZone;
				
			}while( true == xmlDocTree.GoNextSibling() );
		}
		
		xmlDocTree.Go2ParentLevel();
		string strname = xmlDocTree.GetCurElemName();

		outputDocument.dlAppendPage( dlPage );
		
		// delete dlPage;

	} while ( true == xmlDocTree.GoNextSibling() ); // multiple pages?

	xmlDocTree.Go2ParentLevel();

	return outputDocument;
}

void DLxml::traverseXML( xml &xmlDocTree, DLZone *zone )
{
	/* Visit root */
	string strname = xmlDocTree.GetCurElemName(); 

	zone->dlSetZoneOrigin( atoi( xmlDocTree.GetAttrib( "col" ).c_str() ), atoi( xmlDocTree.GetAttrib( "row" ).c_str() ) );
	zone->dlSetZoneHeight( atoi( xmlDocTree.GetAttrib( "height" ).c_str() ) );
	zone->dlSetZoneWidth( atoi( xmlDocTree.GetAttrib( "width" ).c_str() ) );
	
	// modified "zoneID" to "id"
	zone->dlSetZoneID( xmlDocTree.GetAttrib( "id" ).c_str() );
	// zone->dlSetZoneID( xmlDocTree.GetAttrib( "zoneID" ).c_str() );

	map <string, string> dynamicAtt = xmlDocTree.GetDynamicAttrib( 1 );
	map <string, string>::const_iterator iter;
	string zoneAttrName;
	vector <string>::const_iterator str_Iter;
	bool bStaticZoneAttr;
	for(iter = dynamicAtt.begin(); iter != dynamicAtt.end(); iter++)
	{
		bStaticZoneAttr = false;
		zoneAttrName = (*iter).first;
		if(!staticZoneAttributes.empty()) 
		{
			for(str_Iter = staticZoneAttributes.begin(); str_Iter != staticZoneAttributes.end(); str_Iter++)
			{
				if(zoneAttrName == *str_Iter) {
					bStaticZoneAttr = true;
					break;
				}
			}
		}
		if(!bStaticZoneAttr) {
			zone->dlSetTag( (*iter).first, (*iter).second, true );
		}
	}

	/* Insert zones */
	if( xmlDocTree.bHasChild() )
	{
		xmlDocTree.Go2ChildLevel();

		do
		{
			DLZone *dlSubZone = new DLZone();
			traverseXML( xmlDocTree, dlSubZone );
			zone->dlAppendChildZone( dlSubZone );

			// delete dlSubZone;
			
		}while( true == xmlDocTree.GoNextSibling() );

		xmlDocTree.Go2ParentLevel();
	}
}

void DLxml::traverseSubtree( DLZone* zones, int cnt, xml& xmlDocTree )
{
	// insert zone, including subzones
	list <DLZone*>::iterator  Zones_Iter;
	Zones_Iter = zones->begin();
	int zoneCnt = 0;

	for ( Zones_Iter = zones->begin(); Zones_Iter != zones->end(); Zones_Iter++ )
	{
		// visit each child zone
		DLZone* curZone = *(Zones_Iter);

		if( 0 == zoneCnt )
		{
			xmlDocTree.InsertChildElem( "DL_ZONE", 3);
			xmlDocTree.Go2ChildLevel();
			++zoneCnt;
		}
		else
		{
			xmlDocTree.InsertSiblingElem( "DL_ZONE" );
		}

		int startX = curZone->dlGetZoneOrigin().x;
		int startY = curZone->dlGetZoneOrigin().y;
		int width = curZone->dlGetZoneWidth();
		int height = curZone->dlGetZoneHeight();
	
		xmlDocTree.AddDynamicAttrib( curZone->dlGetAllTags().tagMap );
		// xmlDocTree.AddAttrib( "id", curZone->dlGetZoneID().c_str()  );
		xmlDocTree.AddAttrib( "col", startX );
		xmlDocTree.AddAttrib( "row", startY );
		xmlDocTree.AddAttrib( "width", height );
		xmlDocTree.AddAttrib( "height", width );		

		if( true == curZone->dlHasChildZones() )
			traverseSubtree( curZone, 0, xmlDocTree );
	}

	xmlDocTree.Go2ParentLevel();
}

bool DLxml::validatePolygon( DLZone * zone, int pagewidth, int pageheight ) {
	
	int cnt = 0;
	int acceptCnt = 6;
	
	string polyStr = zone->dlGetTag("polygon");
	char * str = new char[polyStr.size()+1];	
	strcpy( str, polyStr.c_str() );
		
	char * token;	
	token = strtok( str, ",;()");
		
	while ( token != NULL )	{
		
		if( cnt % 2 == 0 ) {
			if ( atoi(token) > pagewidth) return false;
		} else {
			if( atoi(token) > pageheight) return false;
		}
		
		cnt++;							
		token = strtok( NULL, ",;()");		
	}
	
	delete[] str;
	
	// at least more than 3 points
	if( cnt >= acceptCnt ) return true;
	else return false;
}